
/*
	The bourgillator (c) Risto Holopainen 2017
	If everything works as it should, the 
	program synthesizes a composition to an audio file. Libsndfile
	must be installed on your system for this to work.
	It doesn't operate in realtime unless you modify it.
	The program isn't supposed to be generally useful; 
	its purpose is to compose a single composition, no more no less.
	Since the composition is already composed this program
	actually serves no other than educational or research purposes.
	
	---
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

*/



#include <cstdio>
#include <cstdlib>
#include <complex>
#include <sndfile.h>
#include "heuns.h"



Vect global_x;	// this global variable holds the output signal x(t)


inline double rnd()
{
	return (double) rand()/RAND_MAX;
}




double heavisides(const Vect &u, int k)
{
	const int N = u.size();
	double s = 0;
	int c = 1 << N;
	for(int i=1, j; i<N; i++, c >>= 1)	// half circle
		{
		j = (k + i) % N;
		if(u[k] > u[j])
			s += c;	
		}
	s /= (1 << N);			// normalize
	return s;
}


Vect steposc(const Vect &u, const Vect &p)
{
	// u: state vector, p: parameters
	// p[0-D-1] = omega_i
	Vect v(u.size());	// return value
	const int D = u.size() / 3;
	static Vect theta(D), a(D), A(D), x(D), xd(D);
	static const double tau = 1.0 / 125000.;
	xd = x; // delay
	
	for(int i=0; i<D; i++)	// unpack
		{
		theta[i] = u[i];
		a[i] = u[i + D];
		A[i] = u[i+2*D];
		x[i] = a[i]*cos(theta[i]);
		global_x[i] = 0.5*(x[i] + xd[i]);
		}

	double g, coupling, A_drift, cc;
	double sigma = p[D];
	double K 	 = p[D+1];
	double delta = p[D+2];	// drift ~ 0.5
	double beta	 = p[D+3];
	
	for(int i=0, j, k, m; i<D; i++)
		{
		j = (i + 1 <  D) ? i+1 : 0;
		k = (i - 1 >= 0) ? i-1 : D-1;
		m = (i + D/2) % D;
		
		A_drift = 1 + delta*(A[j] + A[k] - A[i]);	// m, <--> i
		coupling = sin(theta[i] - theta[j])
			+ sin(theta[i] - theta[k]) 
			+ sin(theta[i] - theta[(i+D-2)%D]) + sin(theta[i] - theta[(i+2)%D]);
		
		cc = (i&3 == 0) ? + -1 : 1;
		coupling += cc * sin(theta[i] - theta[m]);
		
		g = heavisides(A, i);
		theta[i] = sigma*p[i] + A_drift*g - K*A[m]/(0.5 + A[i])*coupling;
		a[i] = g - beta*a[i];		// *a[i];
		A[i] = tau * (x[i]*x[i] - A[i]);
		// repack
		v[i] = theta[i];
		v[i + D] = a[i];
		v[i+2*D] = A[i];
		}
	
	return v;
}


void initvec(Vect &t)
{
	// 3*sz
	double k = 3*M_PI / t.size();
	int sz = t.size() / 3;
	printf("initvec: sz=%d\n", sz);
	srand(37);
	for(int i=0; i<sz; i++)
		{
		t[i]	= k*(i - sz/2);		// theta
		t[i+sz] = 0.0;				// a
		t[i+2*sz] = .01*rnd();		// A
		//printf("%2d: phase %.3f\n", i, t[i]);
		}
}


void init_omega(Vect &w, double step, int sr)
{
	const int N = w.size();
	double frq;
	const double Omega = 2*M_PI/sr/step;
	for(int i=0; i<N; i++)
		{
		frq = 50 * (rnd() - rnd() + rnd());
		w[i] = Omega*frq;
		}
}


void sys_to_smp(const Vect &X, double &l, double &r)
{
	// send global_x to stereo pair
	const int D = X.size();
	l = 0.0;
	r = 0.0;
	for(int j=0, k=1; k<D; j += 2, k += 2)
		{
		l += X[j];
		r += X[k];
		}
	double norm = 1.0 / (5*D + 1);
	l *= norm;
	r *= norm;
}




complex <double> order_parameter(const Vect &u)
{
// expects u: 3*D, uses u[0..D-1]
	const int D = u.size() / 3;
	complex <double> Z = 0.0, I(0.0,1.0);
	
	for(int k=0; k<D; k++)
		{
		Z += exp(I*u[k]);	
		}
	return Z / double(D);
}



int main(int argc, char **argv)
{
	if(argc != 6)
		{
		printf("%s spread coupling drift damping dur.(sec)\n", argv[0]);
		exit(0);
		}

	const int ch = 2;
	SNDFILE *ouf;
	SF_INFO head;
	head.samplerate = 48000;
	head.channels = ch;
	head.format = (SF_FORMAT_WAV | SF_FORMAT_PCM_24);
	ouf = sf_open("bourgillations.wav", SFM_WRITE, &head);
	
	
	const int Dim = 25;
	const int sys_dim = 3*Dim;
	const double stepsize = 0.05;
	heuns HH(stepsize, sys_dim);
	Vect vv(sys_dim), uu(sys_dim);
	Vect omega(Dim);	// D frq-params
	init_omega(omega, stepsize, head.samplerate);
	// other parameters:
	Vect params = omega;
	double sigma = atof(argv[1]);		// freq-spread
	double Coupling = atof(argv[2]);
	double drift = atof(argv[3]);
	double beta = atof(argv[4]);
	unsigned short duration = atoi(argv[5]);
	params.push_back(sigma);
	params.push_back(Coupling);
	params.push_back(drift);	// drift = 0.5		: frq-drift
	params.push_back(beta);		// beta = 0.05--0.5 : damping
	printf("#parameters: %lu\n", params.size());
	for(unsigned int i=0; i<params.size(); i++)
		printf("p%2d = %+.2f %c", i, params[i], (i%3==2) ? '\n' : '\t');

	initvec(vv);
	global_x.resize(Dim);

	const int L = 1<<11;
	double buf[ch*L], ll, rr, maxL = 0.0, maxR = 0.0;
	int T = duration*head.samplerate/L;
	double smp_period = 1.0/head.samplerate;
	
	complex<double> order;
	double time = 0.0;
	bool stats_time = false;
	Vect orderstats;

	for(int i=0; i<T; i++)
		{
		stats_time = (i % 100 == 1);
		order = 0.0;
		for(int j=0, k=0; j<ch*L; j += ch, k++)
			{
			vv = HH.solve(vv, steposc, params);
			// vv---> global_x
			sys_to_smp(global_x, ll, rr);
			buf[j  ] = ll;
			buf[j+1] = rr;
			if(fabs(ll) > fabs(maxL)) maxL = ll;
			if(fabs(rr) > fabs(maxR)) maxR = rr;
			time += smp_period;
			if(stats_time && k<10)
				order += order_parameter(vv);
			}

		sf_write_double(ouf, buf, ch*L);

		if(stats_time)
			{
			order *= 0.1;
			orderstats.push_back(abs(order));
			printf("t = %6.2f s, \tR: %.3f, \tpsi: %.3f\n",
				time, abs(order), arg(order) );
			}
		}

	sf_close(ouf);
	printf("max amp: L %.3f, R %.3f\nfinal state:\n", maxL, maxR);

	double ord_sum = 0.0;	
	for(unsigned int i=0; i<orderstats.size(); i++)
		ord_sum += orderstats[i];

	// final state of A
	for(int i=0; i<Dim; i++)
		{
		printf("%2d a: %.3f\tA: %.3f\n", i, vv[Dim+i], vv[2*Dim+i]);
		}
	
	printf("Order parameter avg. R: %.4f\n", ord_sum / orderstats.size());
	return 0;
}


