
/*
	heuns method for solving ordinary differential equations, 
	second order improved Euler scheme
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
*/

#ifndef HEUNS_H
#define HEUNS_H


#include <valarray>
#include <vector>


using namespace std;

typedef vector<double> Vect ;

typedef valarray <double> realvec;
typedef realvec (*fp) (const realvec &);	// function to integrate
typedef realvec (*fpp) (const realvec &, const realvec &); // w parameters
typedef Vect  (*vfp) (const Vect &);
typedef Vect (*vfpp) (const Vect &, const Vect &);	// 2nd arg: parameters



class heuns
{
	public:
		heuns(double stepsize, const realvec &initializer);
		heuns(double stepsize, int dim);	// for Vect
		realvec solve(const realvec &u, fp func);
		realvec solve(const realvec &u, fpp func, const realvec &param);
		Vect solve(const Vect &u, vfp func);
		Vect solve(const Vect &u, vfpp func, const Vect &parameters);

		

	private:
		unsigned int D;	// dimension of the system
		realvec c1, c2, y;
		double h, hf;
		Vect v1, v2, z;
};

heuns::heuns(double step, const realvec &init)
{
	h = step;
	hf = h/2;
	D = init.size();
	
	c1.resize(D);
	c2.resize(D);
	y.resize(D);
	y = init;
}


heuns::heuns(double step, int dim)
{
// use for Vect.

	h = step;
	hf = h/2;
	D = dim;
	v1.resize(D);
	v2.resize(D);
	z.resize(D);

}


realvec heuns::solve(const realvec &u, fp func)
{
	c1 = func(u);
	y = u + h*c1;
	c2 = func(y);
	return u + hf*(c1 + c2);
}


realvec heuns::solve(const realvec &u, fpp func, const realvec &pp)
{
	c1 = func(u, pp);
	y = u + h*c1;
	c2 = func(y, pp);
	return u + hf*(c1 + c2);
}



Vect heuns::solve(const Vect &X, vfp func)
{
	int N = X.size();
	v1 = func(X);
	z = X;
	for(int i=0; i<N; i++)
		z[i] += h*v1[i];
	v2 = func(z);
	
	for(int i=0; i<N; i++)
		z[i] = X[i] + hf*(v1[i] + v2[i]);
	
	return z;
}





Vect heuns::solve(const Vect &X, vfpp func, const Vect &par)
{
	int N = X.size();
	v1 = func(X, par);
	z = X;
	for(int i=0; i<N; i++)
		z[i] += h*v1[i];
	v2 = func(z, par);
	
	for(int i=0; i<N; i++)
		z[i] = X[i] + hf*(v1[i] + v2[i]);
	
	return z;
}

#endif


